/*******************************************************************************************
*                                                                                          *
*                         ESP-01(S) RealTime Dongle für Zentraleuropa                      *
*                                                                                          *
********************************************************************************************


Anwendungszweck:

  Slave der einen Host (z.B. µC) die aktuelle Zeit per UART mitteilt.
  Diese wird per NTP ermittelt. Der NTP Server stammt aus dem europäischem Pool,
  Zeitzone ist CET bzw. CEST, Sommer-/Winterzeit wird automatisch umgestellt.



Boot:

  1. RealTime kommt aus Reset -> Mist auf der Seriellen Konsole!
  2. RealTime versucht WLAN-Verbindung aufzubauen mit WiFi Manager (wenn keine Verbindung, dann konfiguration via WiFi!)
  3. RealTime macht einmal CR+LF
  4. RealTime meldet "connected" sobald eine Verbindung besteht
  5. RealTime meldet "synced" nach erfolgreicher syncronisation (unendliche retries)
  6. RealTime nimmt ab hier Kommandos entgegen
  


Kommandos:

  Alle Kommandos werden mit CR+LF abgeschlossen, Groß-/Kleinschreibung ist egal.
  8N1 115200 Baud

  Status  - "synced" wenn aktuelle Zeit verfügbar "unsynced" wenn nicht.
  Time    - "HH:MM:SS" im 24h Format für den aktuellen Moment
  Date    - "YYYY:MM:DD"
  Weekday - "W" Tag der Woche (1-7) Sonntag = 1 ... Samstag = 7



Arduino Einstellungen:

  Board: "Generic ESP8266 Module"
  Upload Speed: "115200" (empfohlen, je nach Adapter geht auch schneller)
  CPU Frequency: "160 MHz"
  Crystal Frequency: "26 MHz"
  Flash Size: "1M (no SPIFFS)" für ESP-01S und ESP-01 mit schwarzem PCB
  Flash Size: "512K (no SPIFFS)" für ESP-01 mit blauem PCB
  Flash Mode: "QIO (fast)"
  Flash Frequency: "80MHz"
  Reset Method: "none" für den kleinen roten USB to ESP-01 Adapter
  Debug port: "Disabled"
  Debug Level: "Keine"
  IwIP Variant: "v2 Lower Memory"
  VTables: "Flash"
  Exceptions: "Enabled"
  Builtin Led: "2"
  Erase Flash: "Only Sketch" reicht normalerweise
  Espressif FW: "nonos-sdk 2.2.1 (legacy)"
  SSL Support: "All SSL ciphers (most compatible)"  



Autor:

  Dieser Sketch basiert auf dem Beispiel von Alf Müller (www.bastelgarage.ch) vom Mai 2018 https://youtu.be/vTNIgLoh7OU
  Anpassung auf die Kommandos als UART-Dongle von Robert Steigemann (www.rselec.de) im Oktober 2019

  

*******************************************************************************************/



#include <ESP8266WiFi.h>                    //https://github.com/esp8266/Arduino
#include "WiFiManager.h"                    //https://github.com/tzapu/WiFiManager
#include <Timezone.h>                       //https://github.com/JChristensen/Timezone
#include <TimeLib.h>                        //https://github.com/PaulStoffregen/Time
#include <WiFiUdp.h>


int h, m, s, w, mo, ye, d;                  // variablen für die lokale Zeit

// NTP Serverpool für die Schweiz:
static const char ntpServerName[] = "europe.pool.ntp.org";    //Finde lokale Server unter http://www.pool.ntp.org/zone/ch
const int timeZone = 0;                     // 0 wenn mit <Timezone.h> gearbeitet wird sonst stimmt es nachher nicht
WiFiUDP Udp;
unsigned int localPort = 4321;              // lokaler Port zum Abhören von UDP-Paketen

time_t getNtpTime();
void printDigits(int digits);
void sendNTPpacket(IPAddress &address);

// - Timezone. - //
// Bearbeiten Sie diese Zeilen entsprechend Ihrer Zeitzone und Sommerzeit.
// TimeZone Einstellungen Details https://github.com/JChristensen/Timezone
TimeChangeRule CEST = {"CEST", Last, Sun, Mar, 2, 120};     //Central European Time (Frankfurt, Paris)
TimeChangeRule CET = {"CET ", Last, Sun, Oct, 3, 60};       //Central European Time (Frankfurt, Paris)
Timezone CE(CEST, CET);


void setup() {
  Serial.begin(115200);
  WiFiManager wifiManager;  
   //wifiManager.resetSettings();          //mit diesem befehl kannst die gespeicherten werte löschen              
  wifiManager.setDebugOutput(false);       //suppress serial output
  wifiManager.autoConnect("rs-elec.de");
  Serial.println();
  Serial.println("connected");
  Udp.begin(localPort);
  //Serial.print("lokaler Port: ");
  //Serial.println(Udp.localPort());
  //Serial.println("Warten auf die Synchronisation");
  setSyncProvider(getNtpTime);
  setSyncInterval(2*60*60);                  // Anzahl der Sekunden zwischen dem erneuten Synchronisieren ein. 86400 = 1 Tag
  if (timeStatus() == timeNotSet)            // noch keine NTP Zeit da
  {
    delay(500);
  }
  Serial.println("synced");
}


void loop()
{
  String line_string;
  while(!Serial.available()); 
  if(Serial.available()){
    line_string = Serial.readStringUntil('\n');
  }
  line_string.toLowerCase();
  
  if(line_string.startsWith("status")){
    if (timeStatus() == timeNotSet){
      Serial.println("unsynced");
    }
    else {
      Serial.println("synced");
    }
  }

  if(line_string.startsWith("time")){
    lokaleZeit();
    printDigits(h);
    Serial.print(":");
    printDigits(m);
    Serial.print(":");
    printDigits(s);
    Serial.println();
  }

  if(line_string.startsWith("date")){
    lokaleZeit();
    Serial.print(ye);
    Serial.print(".");
    printDigits(mo);
    Serial.print(".");
    printDigits(d);
    Serial.println();
  }

  if(line_string.startsWith("weekday")){
    lokaleZeit();
    Serial.println(w);
  }
  
}


void lokaleZeit()
{
  time_t tT = now();
  time_t tTlocal = CE.toLocal(tT);
  w = weekday(tTlocal);
  d = day(tTlocal);
  mo = month(tTlocal);
  ye = year(tTlocal);
  h = hour(tTlocal);
  m = minute(tTlocal);
  s = second(tTlocal);
  
}


void printDigits(int digits)
{
  // Utility für digitale Uhrenanzeige: druckt vor Doppelpunkt und führende 0
  //Serial.print(":");
  if (digits < 10)
    Serial.print('0');
  Serial.print(digits);
}

/*-------- NTP code ----------*/

const int NTP_PACKET_SIZE = 48; // NTP-Zeit in den ersten 48 Bytes der Nachricht
byte packetBuffer[NTP_PACKET_SIZE]; //Puffer für eingehende und ausgehende Pakete

time_t getNtpTime()
{
  IPAddress ntpServerIP; // NTP server's ip Adresse

  while (Udp.parsePacket() > 0) ; // alle zuvor empfangenen Pakete verwerfen
  //Serial.println("Transmit NTP Request");
  // einen zufälligen Server aus dem Pool holen
  WiFi.hostByName(ntpServerName, ntpServerIP);
  //Serial.print(ntpServerName);
  //Serial.print(": ");
  //Serial.println(ntpServerIP);
  sendNTPpacket(ntpServerIP);
  uint32_t beginWait = millis();
  while (millis() - beginWait < 15000) {
    int size = Udp.parsePacket();
    if (size >= NTP_PACKET_SIZE) {
      //Serial.println("Receive NTP Response");
      Udp.read(packetBuffer, NTP_PACKET_SIZE);  // Paket in den Puffer einlesen
      unsigned long secsSince1900;
      // vier Bytes ab Position 40 in eine lange Ganzzahl umwandeln
      secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
      secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
      secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
      secsSince1900 |= (unsigned long)packetBuffer[43];
      return secsSince1900 - 2208988800UL + timeZone * SECS_PER_HOUR;
    }
  }
  //Serial.println("keine NTP Antwort");
  return 0; // gibt 0 zurück, wenn die Zeit nicht ermittelt werden kann.
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address)
{
  // alle Bytes im Puffer auf 0 setzen
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialisieren von Werten, die für die Bildung von NTP-Requests benötigt werden.
  // (siehe URL oben für Details zu den Paketen)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // alle NTP-Felder wurden jetzt mit Werten versehen
  // Sie können ein Paket senden, das einen Zeitstempel anfordert.:
  Udp.beginPacket(address, 123); //NTP-Requests sollen auf Port 123 erfolgen
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}
