/**
 *  Modified Serialport Lib to communicate with ESP8266 ESP-01 Module with RealTime FW
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <avr/pgmspace.h>
#include "ESP_RealTime.h"
//#include "Serialport.h"
#include "DS1307.h"


static volatile uint8_t txd_buffer[64];								/**< the 64 Byte transmit buffer */
static volatile uint8_t txd_point_tail = 0;						    /**< pointer offset to the next byte to transmit in buffer */
static volatile uint8_t txd_point_head = 0;							/**< pointer offset to the next position to write in transmit buffer */

static volatile uint8_t rxd_buffer[256];							/**< the 256 Byte receive buffer */
static volatile uint8_t rxd_point_tail = 0;						    /**< pointer offset to the last read byte in buffer */
static volatile uint8_t rxd_point_head = 0;							/**< pointer offset to the next position to write in receive buffer */


//int8_t hours = -1;
//int8_t minutes = -1;
//int8_t seconds = -1;
//
//int16_t year = -1;
//int8_t month = -1;
//int8_t day = -1;                                             
//int8_t weekday = -1;                                         // Tag der Woche (1-7) Sonntag = 1 ... Samstag = 7


static rtc_time_t nowTime;
static rtc_date_t nowDate;


// This function add an array to the transmit buffer and control the DRE interrupt.
void espport_write(uint8_t *anArray, uint8_t arrayLength)
{	
	uint8_t counter;												// the count variable over the array length
	
	for (counter=0;counter<arrayLength;counter++)					// for each byte in the array
	{				
		ESPPORT_UART.CTRLA &= 0b11111100;						// deactivate the DRE (data register empty) interrupt 			
		*(txd_buffer+txd_point_head) = *(anArray+counter);			// add a byte from the array to the transmit buffer		
		txd_point_head = 0b00111111 & (txd_point_head + 1);			// increment the last byte "pointer" and mask it for 64 to match the buffer size
		ESPPORT_UART.CTRLA |= 0b00000001;						// activate the DRE interrupt 		
	}			
}


// This interrupt is executed, when the data register is empty. It put the next byte into the data register of the UART to transmit it.
// This interrupt is only enabled, if the transmit buffer hold at least one byte to transmit!
ISR(ESPPORT_DRE_vect)
{		
	ESPPORT_UART.DATA = txd_buffer[txd_point_tail];				// put the next byte into the data register	
	txd_point_tail = 0b00111111 & (txd_point_tail + 1);				// increment the next byte pointer and mask it for 64 to match the buffer size
	
	// if the last byte is transmitting	
	if (txd_point_tail==txd_point_head)
	{
		ESPPORT_UART.CTRLA &= 0b11111100;						// deactivate the DRE interrupt 		
	}		
}


// This function checks how many space are in the transmit buffer.
uint8_t espport_checkSpace(void)
{
	if ((ESPPORT_UART.CTRLA & 0b00000011) == 0)					// there is no data to transmit, because the interrupt is off
	{
		return 64;													// so simply return 64 bytes are free
	} else if (txd_point_head == txd_point_tail)					// buffer is complete full!
	{
		return 0;
	} else															// calculate free buffer
	{
		return 64 - ((txd_point_head - txd_point_tail) & 0b00111111);			
	}
}


// This interrupt is executed, when a byte is received. It put it on the head of the receive buffer and increment the head pointer.
ISR(ESPPORT_RXC_vect)
{	
	rxd_buffer[rxd_point_head] = ESPPORT_UART.DATA;				// copy byte from data register to receive buffer			
	rxd_point_head++;												// increment the receive head pointer	
}



// This function initializes the USART to 8N1 with given Baudrate and enables the RXC interrupt.
uint8_t espport_init(uint32_t baudrate)
{		
	uint8_t  clk2_bm;		// double speed bit mask, set to USART_CLK2X_bm for double speed
	int8_t   bscale;		// -7 to 7
	uint32_t bsel;			// 1 to 4095
	uint32_t temp;
		
	if ((F_CPU / baudrate) <= 32 )	// Baudrate is high compared to CPU Frequency, so use double speed
	{
		clk2_bm = USART_CLK2X_bm;
		
		// check the negative bscale values -7 to -1
		for (bscale=7; bscale > 0; bscale--)	// 7 to 1
		{
			temp = baudrate<<3;
			bsel = ((F_CPU<<bscale)+(temp>>1))/temp - (1<<bscale);
			if (bsel <= 4095)
			{
				bscale *= -1;
				break;
			}
		}
		
		if (bsel > 4095)	// no match so far
		{
			// check the positive bscale values 0 to 7
			for (bscale=0; bscale <= 7; bscale++)	// 0 to 7
			{
				temp = baudrate<<(3+bscale);
				bsel = (F_CPU+(temp>>1))/temp - 1;
				if (bsel <= 4095)
				{
					break;
				}
			}
		}
	}
	else							// use normal speed with better sampling
	{
		clk2_bm = 0;
		
		// check the negative bscale values -7 to -1
		for (bscale=7; bscale > 0; bscale--)	// 7 to 1
		{
			temp = baudrate<<4;
			bsel = ((F_CPU<<bscale)+(temp>>1))/temp - (1<<bscale);
			if (bsel <= 4095)
			{
				bscale *= -1;
				break;
			}
		}
		
		if (bsel > 4095)	// no match so far
		{
			// check the positive bscale values 0 to 7
			for (bscale=0; bscale <= 7; bscale++)	// 0 to 7
			{
				temp = baudrate<<(4+bscale);
				bsel = (F_CPU+(temp>>1))/temp - 1;
				if (bsel <= 4095)
				{
					break;
				}
			}
		}
	}
	
	if ((bsel > 0) && (bsel <= 4095))	// match found
	{
		ESPPORT_UART.BAUDCTRLB = (((uint8_t)bscale<<4) | (uint8_t)(bsel/256));	// map values to baudrate registers
		ESPPORT_UART.BAUDCTRLA = (uint8_t)(bsel % 256);								
		ESPPORT_UART.CTRLA = USART_RXCINTLVL_LO_gc;							// set interrupt level for receive interrupt to high 
		ESPPORT_UART.CTRLC = USART_CHSIZE_8BIT_gc;							// use 8 Bit Modus	
		ESPPORT_UART.CTRLB = USART_TXEN_bm | USART_RXEN_bm | clk2_bm;		// enable receiver and transmitter and double speed if needed		
		// set pin directions
        ESPPORT_TXD_TO_OUTPUT;
        ESPPORT_RXD_TO_INPUT;
        
        // and default states of control signals, Boot is unused yet
        ESPPORT_EN_LOW;
        ESPPORT_EN_TO_OUTPUT;
        ESPPORT_RST_HIGH;
        ESPPORT_RST_TO_OUTPUT;        
        return 0;		
	}
	else		// selected baudrate is impossible
	{
		return 1;
	}
}


// This function gives you all received data to the next CR (carriage return) and LF in string format.
// To get the string, it is necessary that your array is big enough to hold all data to the CR.
// If is not (maxLength reached, before CR is) this function return 0 and data is lost.
uint8_t espport_getLastString(uint8_t *anArray, uint8_t maxLength)
{	
	uint8_t counter;												// the count variable over the array length
	uint8_t cr_is_there = 0;										// becomes true if we found a CR
	
	// run through all filled buffer bytes and search for a CR
	for (counter=rxd_point_tail;counter!=rxd_point_head;counter++)
	{						
		if (rxd_buffer[counter]==13) {								// 13 == CR (carriage return)
			cr_is_there = 1;										// CR found!
			break;													// don't search for a second one
		}		
	}
	
	// if we have a CR
	if (cr_is_there)
	{		
		counter = 0;												// clear counter		
				
		while (rxd_buffer[rxd_point_tail]!=13)						// until we reach the CR
		{			
			*(anArray+counter) = rxd_buffer[rxd_point_tail];		// copy data to array			
			rxd_point_tail++;										// increment tail
			counter++;												// and the counter
			
			// If the array is full and CR is not reached!
			if ((counter == (maxLength-1)) && (rxd_buffer[rxd_point_tail]!=13))
			{
				*(anArray+counter) = 0;								// add the '\0' to the end of string
				rxd_point_tail++;									// also need to increment the tail				
				return 0;											// return 0 (change this if you want)
			}									
		}
				
		*(anArray+counter) = 0;										// add the '\0' to the end of string	
		rxd_point_tail+=2;											// need to increment the tail two times, because we assume that we have CR+LF, but we don't work with LF
		
		return counter+1;											// return the received characters + '\0'
				
	}
	else															// we haven't found a CR
	{																
		return 0;													// nothing received to return		
	}	
}

// printf for uart
void espport_printf(const char* __fmt, ...)
{
	char aString[64];
	va_list argumentlist;
	va_start(argumentlist, __fmt);
	vsprintf(aString, __fmt, argumentlist);
	va_end(argumentlist);
	while(strlen(aString)>espport_checkSpace());
	espport_write((uint8_t *)aString, strlen(aString));
}

// Simple print a string
void espport_print(char* aString)
{
	while(strlen(aString)>espport_checkSpace());
	espport_write((uint8_t *)aString, strlen(aString));
}



///////////////////////////////////////////////////////////////
/// Here are the high-Level functions for ESP RealTime      ///
///////////////////////////////////////////////////////////////


int8_t espport_boot_and_connect(void)
{
    uint8_t myString[16] = "";
    uint16_t timeout = 10000;        // timeout in ms
    uint8_t stringlength = 0; 
        
    ESPPORT_RST_LOW;
    ESPPORT_EN_HIGH;
    ESPPORT_RST_HIGH;
    //The ESP-01 with RealTime should now boot and connect
    do 
    {
        if (timeout == 0)
        {
            return -1;
        }
        _delay_ms(1);
        timeout--;
        stringlength = espport_getLastString(myString, 16);
    } while (strcasecmp_P((char *)myString, PSTR("synced")) != 0);
    return 0;
}


void espport_disconnect(void)
{
    ESPPORT_EN_LOW;
}


int8_t espport_get_time(void)
{
    uint8_t myString[16] = "";
    uint16_t timeout = 100;        // timeout in ms
    
    espport_print("time\r\n");
    do
    {
        if (timeout == 0)
        {
            return -1;
        }
        _delay_ms(1);
        timeout--;       
    } while (espport_getLastString(myString, 16) != 9); //HH:MM:SS\0 should be the only response
    
    nowTime.hours   = (myString[1]-'0') + 10 * (myString[0]-'0');
    nowTime.minutes = (myString[4]-'0') + 10 * (myString[3]-'0');
    nowTime.seconds = (myString[7]-'0') + 10 * (myString[6]-'0');
    
    return 0;
}



int8_t espport_get_date(void)
{
    uint8_t myString[16] = "";
    uint16_t timeout = 100;        // timeout in ms
    
    espport_print("date\r\n");
    do
    {
        if (timeout == 0)
        {
            return -1;
        }
        _delay_ms(1);
        timeout--;
    } while (espport_getLastString(myString, 16) != 11); //YYYY:MM:DD\0 should be the only response
    
    nowDate.year  = (myString[3]-'0') + 10 * (myString[2]-'0') + 2000;
    nowDate.month = (myString[6]-'0') + 10 * (myString[5]-'0');
    nowDate.day   = (myString[9]-'0') + 10 * (myString[8]-'0');
    
    return 0;
}


int8_t espport_get_weekday(void)
{
    uint8_t myString[16] = "";
    uint16_t timeout = 100;        // timeout in ms
    
    espport_print("weekday\r\n");
    do
    {
        if (timeout == 0)
        {
            return -1;
        }
        _delay_ms(1);
        timeout--;
    } while (espport_getLastString(myString, 16) != 2); //W\0 should be the only response
    
    nowDate.weekday = myString[0]-'0';
    
    return 0;
}


int8_t espport_update(void)
{
    int8_t retval = 0;
    
    retval = espport_boot_and_connect();
    if (retval == 0)
    {
        retval = espport_get_time();
        if (retval == 0)
        {
            rtc_set_time(nowTime);
        }
        else
        {
            return retval;  // something goes wrong
        }        
        retval += espport_get_date();       // get date
        retval += espport_get_weekday();
        retval += espport_get_time();       // get time again
        if (retval == 0)
        {
            if (nowTime.hours==0 && nowTime.minutes==0 && nowTime.seconds==0)   
            {
                // a new day has started right now, we are unsure if date and weekday are requested yesterday or today
                // so we request them once again!
                retval += espport_get_date();       // get date
                retval += espport_get_weekday();
            }
        }
        else
        {
            return retval;   // something goes wrong
        }
        if (retval == 0)
        {
            rtc_set_date(nowDate);
        }
        else
        {
            return retval;   // something goes wrong
        }
    }
    
    espport_disconnect();
    return retval;        
}

