/*************************************************************************
* Title:    I2C master library using hardware TWI interface
* Author:   Robert Steigemann for ATXmega, old: Peter Fleury <pfleury@gmx.ch>  http://jump.to/fleury
* File:     $Id: twimaster.c,v 1.4 2015/01/17 12:16:05 peter Exp $
* Software: AVR-GCC 3.4.3 / avr-libc 1.2.3
* Target:   any AVR device with hardware TWI 
* Usage:    API compatible with I2C Software Library i2cmaster.h
**************************************************************************/


#include <avr/io.h>
#include <util/delay.h>
#include "i2cmaster.h"



/* I2C clock in Hz */
#define SCL_CLOCK  100000L


/*************************************************************************
 Initialization of the I2C bus interface. Need to be called only once
*************************************************************************/
void i2c_init(void)
{
  /* initialize TWI clock: 100 kHz clock, TWPS = 0 => prescaler = 1 */
  TWIPORT.MASTER.CTRLA = TWI_MASTER_ENABLE_bm;
  TWIPORT.MASTER.BAUD = F_CPU/(2*SCL_CLOCK)-5;
  TWIPORT.MASTER.STATUS |= TWI_MASTER_BUSSTATE_IDLE_gc;
}/* i2c_init */


/*************************************************************************	
  Issues a start condition and sends address and transfer direction.
  return 0 = device accessible, 1= failed to access device
*************************************************************************/
unsigned char i2c_start(unsigned char address)
{
    TWIPORT.MASTER.ADDR = address;
    while(!((TWIPORT.MASTER.STATUS & TWI_MASTER_WIF_bm) || (TWIPORT.MASTER.STATUS & TWI_MASTER_RIF_bm)));

	return 0;
}/* i2c_start */


/*************************************************************************
 Issues a repeated start condition and sends address and transfer direction 

 Input:   address and transfer direction of I2C device
 
 Return:  0 device accessible
          1 failed to access device
*************************************************************************/
unsigned char i2c_rep_start(unsigned char address)
{
    return i2c_start( address );

}/* i2c_rep_start */


/*************************************************************************
 Terminates the data transfer and releases the I2C bus
*************************************************************************/
void i2c_stop(void)
{
    TWIPORT.MASTER.CTRLC = TWI_MASTER_CMD_STOP_gc;
}/* i2c_stop */



/*************************************************************************
  Send one byte to I2C device
  
  Input:    byte to be transfered
  Return:   0 write successful 
            1 write failed
*************************************************************************/
unsigned char i2c_write( unsigned char data )
{	
    TWIPORT.MASTER.DATA = data;
    while(!(TWIPORT.MASTER.STATUS & TWI_MASTER_WIF_bm));
    
	return 0;

}/* i2c_write */


/*************************************************************************
 Read one byte from the I2C device, request more data from device 
 
 Return:  byte read from I2C device
*************************************************************************/
unsigned char i2c_readAck(void)
{
	uint8_t data = 0;
    
    while(!(TWIPORT.MASTER.STATUS & TWI_MASTER_RIF_bm));
	data = TWIPORT.MASTER.DATA;
    
    TWIPORT.MASTER.CTRLC = TWI_MASTER_CMD_RECVTRANS_gc;

    return data;
}/* i2c_readAck */


/*************************************************************************
 Read one byte from the I2C device, read is followed by a stop condition 
 
 Return:  byte read from I2C device
*************************************************************************/
unsigned char i2c_readNak(void)
{
	uint8_t data = 0;
	
	while(!(TWIPORT.MASTER.STATUS & TWI_MASTER_RIF_bm));
	data = TWIPORT.MASTER.DATA;
	
	TWIPORT.MASTER.CTRLC = TWI_MASTER_ACKACT_bm | TWI_MASTER_CMD_STOP_gc;

	return data;
}/* i2c_readNak */
