/*
 * tube_ctrl.c
 *
 */ 

#include <avr/io.h>
#include <stdbool.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "tube_ctrl.h"


typedef enum
{
    PH_TO_FRONT,
    PH_TO_BACK,
    PH_TO_TRAGET
} phase;


// globals for slot machine (sm)
// mt = minutes tens 00:X0
// mo = minutes ones 00:0X
// ht = hours tens   X0:00
// ho = hours ones   0X:00
// without postfix = actual shown number, -1 = off
// postfix _t      = target to be reached by slot machine

static volatile int8_t mt, mt_t;
static volatile int8_t mo, mo_t;
static volatile int8_t ht, ht_t;
static volatile int8_t ho, ho_t;
static bool colon_state;

// every SM_TYPE needs to implement its version of the start_slotmachine function,
// it should use the TUBE_TIMER to reach the target values in their specific way 
static void start_slotmachine(int8_t target_ht, int8_t target_ho, int8_t target_mt, int8_t target_mo, uint16_t speed);


static int8_t pos_to_number(int8_t position)
{
    // Anordnung der Ziffern in einer CD66A Tube von Hinten (Index 0) nach Vorne (Index 9)
    const int8_t pos_to_number_table[10] = {0, 1, 2, 3, 9, 4, 5, 8, 7, 6};
    
    if (position >= 0 && position <= 9) // position in range
    {
        return (pos_to_number_table[position]);
    } 
    else // position is not in range
    {
        return -1; // assume digit is off
    }    
}

static int8_t number_to_pos(int8_t number)
{
    const int8_t number_to_pos_table[10] = {0, 1, 2, 3, 5, 6, 9, 8, 7, 4};
        
    if (number >= 0 && number <= 9) // number exists in tube
    {
        return (number_to_pos_table[number]);
    } 
    else // number doesn't exist in tube
    {
        return -1; // assume digit is off
    }        
}


void tube_hv(bool enable)
{
    if (enable)
    {
        PORTR_OUTSET = PIN0_bm;          // turn HV_EN on
         _delay_ms(10);                     // how long is the rampup time for voltage?
    }
    else
    {
        PORTR_OUTCLR = PIN0_bm;          // turn HV_EN off
    }
}


// control colon between hours and minutes
void tube_set_colon(bool enable)
{
    if (enable)
    {
        PORTH_OUTSET = PIN3_bm;
    }
    else
    {
        PORTH_OUTCLR = PIN3_bm;
    }
    colon_state = enable;
}


void tube_toggle_colon(void)
{
    tube_set_colon(!colon_state);
}


bool tube_ht_on(void)
{
    if (ht >= 0 && ht <= 9)
    {
        return true;
    } 
    else
    {
        return false;
    }
}


// control decimal point of minutes ones (fourth digit from left = rightmost digit)
static void show_dp_mo(bool enable)
{
    if (enable)
    {
        PORTJ_OUTSET = PIN7_bm;
    }
    else
    {
        PORTJ_OUTCLR = PIN7_bm;
    }
}


// control decimal point of minutes tens (third digit from left)
static void show_dp_mt(bool enable)
{
    if (enable)
    {
        PORTH_OUTSET = PIN4_bm;
    }
    else
    {
        PORTH_OUTCLR = PIN4_bm;
    }
}


// control decimal point of hours ones (second digit from left)
static void show_dp_ho(bool enable)
{
    if (enable)
    {
        PORTD_OUTSET = PIN5_bm;
    }
    else
    {
        PORTD_OUTCLR = PIN5_bm;
    }
}


// control decimal point of hours tens (first digit from left)
static void show_dp_ht(bool enable)
{
    if (enable)
    {
        PORTB_OUTSET = PIN1_bm;
    }
    else
    {
        PORTB_OUTCLR = PIN1_bm;
    }
}


// change or show digit of minutes ones (fourth digit from left = rightmost digit)
static void show_digit_mo(int8_t number)
{
    // turn off all numbers of this digit
    PORTQ_OUTCLR = 0b00001100;
    PORTK_OUTCLR = 0b11111111;
    
    // turn on correct number
    switch (number)
    {
    case 0:
        PORTQ_OUTSET = PIN3_bm;
    	break;
        
    case 1:
        PORTK_OUTSET = PIN0_bm;
        break;
        
    case 2:
        PORTK_OUTSET = PIN1_bm;
        break;
        
    case 3:
        PORTK_OUTSET = PIN2_bm;
        break;
    
    case 4:
        PORTK_OUTSET = PIN3_bm;
        break;
    
    case 5:
        PORTK_OUTSET = PIN4_bm;
        break;
    
    case 6:
        PORTK_OUTSET = PIN5_bm;
        break;
    
    case 7:
        PORTK_OUTSET = PIN6_bm;
        break;
    
    case 8:
        PORTK_OUTSET = PIN7_bm;
        break;
    
    case 9:
        PORTQ_OUTSET = PIN2_bm;
        break;
    
    default:
        // digit off
        mo = -1;
        return;        
    }
    
    mo = number;    
}


// change or show digit of minutes tens (third digit from left)
static void show_digit_mt(int8_t number)
{
    // turn off all numbers of this digit
    PORTJ_OUTCLR = 0b01111111;
    PORTH_OUTCLR = 0b11100000;
    
    // turn on correct number
    switch (number)
    {
        case 0:
            PORTJ_OUTSET = PIN6_bm;
            break;
        
        case 1:
            PORTH_OUTSET = PIN5_bm;
            break;
        
        case 2:
            PORTH_OUTSET = PIN6_bm;
            break;
        
        case 3:
            PORTH_OUTSET = PIN7_bm;
            break;
        
        case 4:
            PORTJ_OUTSET = PIN0_bm;
            break;
        
        case 5:
            PORTJ_OUTSET = PIN1_bm;
            break;
        
        case 6:
            PORTJ_OUTSET = PIN2_bm;
            break;
        
        case 7:
            PORTJ_OUTSET = PIN3_bm;
            break;
        
        case 8:
            PORTJ_OUTSET = PIN4_bm;
            break;
        
        case 9:
            PORTJ_OUTSET = PIN5_bm;
            break;
        
        default:
            // digit off
            mt = -1;
            return;
    }
    
    mt = number;
}


// change or show digit of hours ones (second digit from left)
static void show_digit_ho(int8_t number)
{
    // turn off all numbers of this digit
    PORTE_OUTCLR = 0b00111111;
    PORTD_OUTCLR = 0b11000000;
    PORTF_OUTCLR = 0b00110000;
    
    // turn on correct number
    switch (number)
    {
        case 0:
            PORTF_OUTSET = PIN5_bm;
            break;
        
        case 1:
            PORTD_OUTSET = PIN6_bm;
            break;
        
        case 2:
            PORTD_OUTSET = PIN7_bm;
            break;
        
        case 3:
            PORTE_OUTSET = PIN0_bm;
            break;
        
        case 4:
            PORTE_OUTSET = PIN1_bm;
            break;
        
        case 5:
            PORTE_OUTSET = PIN2_bm;
            break;
        
        case 6:
            PORTE_OUTSET = PIN3_bm;
            break;
        
        case 7:
            PORTE_OUTSET = PIN4_bm;
            break;
        
        case 8:
            PORTE_OUTSET = PIN5_bm;
            break;
        
        case 9:
            PORTF_OUTSET = PIN4_bm;
            break;
        
        default:
            // digit off
            ho = -1;
            return;
    }
    
    ho = number;
}


// change or show digit of hours tens (first digit from left)
static void show_digit_ht(int8_t number)
{
    // turn off all numbers of this digit
    PORTB_OUTCLR = 0b00011100;
    PORTC_OUTCLR = 0b00110000;
    PORTD_OUTCLR = 0b00011111;
    
    // turn on correct number
    switch (number)
    {
        case 0:
            PORTD_OUTSET = PIN4_bm;
            break;
        
        case 1:
            PORTB_OUTSET = PIN2_bm;
            break;
        
        case 2:
            PORTB_OUTSET = PIN3_bm;
            break;
        
        case 3:
            PORTB_OUTSET = PIN4_bm;
            break;
        
        case 4:
            PORTC_OUTSET = PIN4_bm;
            break;
        
        case 5:
            PORTC_OUTSET = PIN5_bm;
            break;
        
        case 6:
            PORTD_OUTSET = PIN0_bm;
            break;
        
        case 7:
            PORTD_OUTSET = PIN1_bm;
            break;
        
        case 8:
            PORTD_OUTSET = PIN2_bm;
            break;
        
        case 9:
            PORTD_OUTSET = PIN3_bm;
            break;
        
        default:
            // digit off
            ht = -1;
            return;
    }
    
    ht = number;
}


// initialize all io pins at startup, clear display
void tube_init(void)
{
    // turn HV off and pin as output
    PORTR_OUTCLR = PIN0_bm;
    PORTR_DIRSET = PIN0_bm;
    
    // turn everything off
    tube_set_colon(false);
    show_dp_mo(false);
    show_dp_mt(false);
    show_dp_ho(false);
    show_dp_ht(false);
    show_digit_mo(-1);
    show_digit_mt(-1);
    show_digit_ho(-1);
    show_digit_ht(-1);    
    
    // configure all digit pins as output
    PORTB_DIRSET = 0b00011110;
    PORTC_DIRSET = 0b00110000;
    PORTD_DIRSET = 0b11111111;
    PORTE_DIRSET = 0b00111111;
    PORTF_DIRSET = 0b00110000;
    PORTH_DIRSET = 0b11111000;
    PORTJ_DIRSET = 0b11111111;
    PORTK_DIRSET = 0b11111111;
    PORTQ_DIRSET = 0b00001100;
}


// simple test of all tube digits
void tube_test(void)
{
    uint8_t i = 0;
    
    tube_hv(true);   
    
    for(i=0; i<=9; i++)
    {
        show_digit_mo(i);
        show_digit_mt(i);
        show_digit_ho(i);
        show_digit_ht(i);
        
        tube_set_colon((i%2)==0);        
        _delay_ms(1000);        
    }
    show_dp_mo(true);
    show_dp_mt(true);
    show_dp_ho(true);
    show_dp_ht(true);
    _delay_ms(1000);
    
    // turn everything off
    tube_set_colon(false);
    show_dp_mo(false);
    show_dp_mt(false);
    show_dp_ho(false);
    show_dp_ht(false);
    show_digit_mo(-1);
    show_digit_mt(-1);
    show_digit_ho(-1);
    show_digit_ht(-1);
    
    tube_hv(false);
}


void show_digit(uint8_t digit, int8_t number)
{
    switch (digit)
    {
    case 0:
        show_digit_ht(number);
    	break;
        
    case 1:
        show_digit_ho(number);
        break;
        
    case 2:
        show_digit_mt(number);
        break;
        
    case 3:
        show_digit_mo(number);
        break;
        
    default:
        // show nothing
        break;       
    }
}


void tube_show_time(uint8_t hours, uint8_t minutes)
{
    if (hours < 10)
    {
        start_slotmachine(-1, hours, minutes/10, minutes%10, SM_SPEED_MS);
    }
    else
    {
        start_slotmachine(hours/10, hours%10, minutes/10, minutes%10, SM_SPEED_MS);
    }
}


#ifdef SM_TYPE_FRONT_BACK_TARGET

    static volatile phase sm_phase = PH_TO_FRONT;

    static void start_slotmachine(int8_t target_ht, int8_t target_ho, int8_t target_mt, int8_t target_mo, uint16_t speed)
    {
        TUBE_TIMER.CTRLA = TC_CLKSEL_OFF_gc;         // stop timer
    
        ht_t = target_ht;                            // store new targets for interrupt routine
        ho_t = target_ho;
        mt_t = target_mt;
        mo_t = target_mo;
    
        sm_phase = PH_TO_FRONT;
            
        TUBE_TIMER.PER = (uint16_t)(((F_CPU/1000)*(uint32_t)speed)/1024 - 1);
        TUBE_TIMER.CNT = TUBE_TIMER.PER - 1;         // prepare timer to raise interrupt directly 
        TUBE_TIMER.INTCTRLA = TC_OVFINTLVL_MED_gc;   // raise low priority interrupt on overflow
        TUBE_TIMER.CTRLA = TC_CLKSEL_DIV1024_gc;     // start timer    
    }

    ISR(TUBE_OVF_vect)        
    {
        // postfix _p = position not the number, see pos_to_number lookup table
        int8_t ht_p, ho_p, mt_p, mo_p;
    
        ht_p = number_to_pos(ht);
        ho_p = number_to_pos(ho);
        mt_p = number_to_pos(mt);
        mo_p = number_to_pos(mo);
    
    
        // first state of slot machine, bring all digits to the front
        if (sm_phase == PH_TO_FRONT)
        {
            if (ht_p < 9)
            {
                show_digit_ht(pos_to_number(ht_p+1));
            }
            if (ho_p < 9)
            {
                show_digit_ho(pos_to_number(ho_p+1));
            }
            if (mt_p < 9)
            {
                show_digit_mt(pos_to_number(mt_p+1));
            }
            if (mo_p < 9)
            {
                show_digit_mo(pos_to_number(mo_p+1));
            }
            if (ht_p==9 && ho_p==9 && mt_p==9 && mo_p==9)
            {
                sm_phase = PH_TO_BACK;
            }
        }
    
        if (sm_phase == PH_TO_BACK)
        {
            if (ht_p > 0)
            {
                show_digit_ht(pos_to_number(ht_p-1));
            }
            if (ho_p > 0)
            {
                show_digit_ho(pos_to_number(ho_p-1));
            }
            if (mt_p > 0)
            {
                show_digit_mt(pos_to_number(mt_p-1));
            }
            if (mo_p > 0)
            {
                show_digit_mo(pos_to_number(mo_p-1));
            }
            if (ht_p==0 && ho_p==0 && mt_p==0 && mo_p==0)
            {
                sm_phase = PH_TO_TRAGET;
            }
        }
    
        if (sm_phase == PH_TO_TRAGET)
        {
            if (ht_t == -1)
            {
                show_digit_ht(-1);
            }
            else if (ht != ht_t)
            {
                show_digit_ht(pos_to_number(ht_p+1));
            }
            if (ho != ho_t)
            {
                show_digit_ho(pos_to_number(ho_p+1));
            }
            if (mt != mt_t)
            {
                show_digit_mt(pos_to_number(mt_p+1));
            }
            if (mo != mo_t)
            {
                show_digit_mo(pos_to_number(mo_p+1));
            }
        }
    }

#endif


#ifdef SM_TYPE_ROLL_BACK
   
    static volatile bool ht_must_run_once;
   
    // go one digit back and clip to front
    static int8_t goBack(int8_t number)
    {
        int8_t pos;
        pos = number_to_pos(number);
        if (pos <= 0) // end or off, start at front
        {
            return (pos_to_number(9));
        } 
        else
        {
            return (pos_to_number(pos-1));
        }       
    }        

    static void start_slotmachine(int8_t target_ht, int8_t target_ho, int8_t target_mt, int8_t target_mo, uint16_t speed)
    {
        TUBE_TIMER.CTRLA = TC_CLKSEL_OFF_gc;         // stop timer
    
        ht_t = target_ht;                            // store new targets for interrupt routine
        ho_t = target_ho;
        mt_t = target_mt;
        mo_t = target_mo;
        
        ht_must_run_once = true;
        
        TUBE_TIMER.PER = (uint16_t)(((F_CPU/1000)*(uint32_t)speed)/1024 - 1);
        TUBE_TIMER.CNT = TUBE_TIMER.PER - 1;         // prepare timer to raise interrupt directly
        TUBE_TIMER.INTCTRLA = TC_OVFINTLVL_MED_gc;   // raise low priority interrupt on overflow
        TUBE_TIMER.CTRLA = TC_CLKSEL_DIV1024_gc;     // start timer
    }

    ISR(TUBE_OVF_vect)
    {
        if (mo != mo_t)
        {
            show_digit_mo(goBack(mo));
        }
        
        if (mt != mt_t)
        {
            show_digit_mt(goBack(mt));
        }       
        
        if (ht_t == -1 && ht != -1)     // 0 to 9 o'clock and ht still shows something
        {           
            if (number_to_pos(ht) != 0)
            {
                show_digit_ht(goBack(ht));
            }
            else
            {
                show_digit_ht(-1);  // turn hours tens off
            }                                
        }
         
        if (ht_t != -1)                // 10 to 23 o'clock
        {            
            if (ho != ho_t && ht_must_run_once) // hour has changed and ht not moved yet
            {
                ht_must_run_once = false;
                show_digit_ht(goBack(ht));  // change also ht in the first cycle, even if ht is correct
            }
            else if (ht != ht_t)
            {
                show_digit_ht(goBack(ht));
            }
        }
        
        if (ho != ho_t)
        {
            show_digit_ho(goBack(ho));
        }                
    } // END ISR

#endif // END #ifdef SM_TYPE_ROLL_BACK

#ifdef SM_TYPE_ALL_BACK_TARGET          // All digits go to back, jump to front and go back to target

    static volatile phase ht_phase, ho_phase, mt_phase, mo_phase;

    static void start_slotmachine(int8_t target_ht, int8_t target_ho, int8_t target_mt, int8_t target_mo, uint16_t speed)
    {
        TUBE_TIMER.CTRLA = TC_CLKSEL_OFF_gc;         // stop timer
    
        ht_t = target_ht;                            // store new targets for interrupt routine
        ho_t = target_ho;
        mt_t = target_mt;
        mo_t = target_mo;
    
        ht_phase = PH_TO_BACK;
        ho_phase = PH_TO_BACK;
        mt_phase = PH_TO_BACK;
        mo_phase = PH_TO_BACK;
    
        TUBE_TIMER.PER = (uint16_t)(((F_CPU/1000)*(uint32_t)speed)/1024 - 1);
        TUBE_TIMER.CNT = TUBE_TIMER.PER - 1;         // prepare timer to raise interrupt directly
        TUBE_TIMER.INTCTRLA = TC_OVFINTLVL_MED_gc;   // raise low priority interrupt on overflow
        TUBE_TIMER.CTRLA = TC_CLKSEL_DIV1024_gc;     // start timer
    }

    ISR(TUBE_OVF_vect)
    {
        // postfix _p = position not the number, see pos_to_number lookup table
        int8_t ht_p, ho_p, mt_p, mo_p;
        
        ht_p = number_to_pos(ht);
        ho_p = number_to_pos(ho);
        mt_p = number_to_pos(mt);
        mo_p = number_to_pos(mo);
        
        
        if (mo_phase == PH_TO_BACK)
        {
            if (mo_p <= 0)  // back reached or off
            {
                mo_phase = PH_TO_TRAGET;               // switch phase for next cycle
                show_digit_mo(pos_to_number(9));        // show on front
            } 
            else
            {
                show_digit_mo(pos_to_number(mo_p-1));   // go back
            }
        } 
        else // go to target
        {
            if (mo != mo_t) // target not reached
            {
                show_digit_mo(pos_to_number(mo_p-1));   // go back
            }
        }
        
        if (mt_phase == PH_TO_BACK)
        {
            if (mt_p <= 0)  // back reached or off
            {
                mt_phase = PH_TO_TRAGET;               // switch phase for next cycle
                show_digit_mt(pos_to_number(9));        // show on front
            }
            else
            {
                show_digit_mt(pos_to_number(mt_p-1));   // go back
            }
        }
        else // go to target
        {
            if (mt != mt_t) // target not reached
            {
                show_digit_mt(pos_to_number(mt_p-1));   // go back
            }
        }
        
        if (ho_phase == PH_TO_BACK)
        {
            if (ho_p <= 0)  // back reached or off
            {
                ho_phase = PH_TO_TRAGET;               // switch phase for next cycle
                show_digit_ho(pos_to_number(9));        // show on front
            }
            else
            {
                show_digit_ho(pos_to_number(ho_p-1));   // go back
            }
        }
        else // go to target
        {
            if (ho != ho_t) // target not reached
            {
                show_digit_ho(pos_to_number(ho_p-1));   // go back
            }
        }
        
        if (ht_t >= 0 && ht >= 0)  // hours ten are lit and should lit
        {
            if (ht_phase == PH_TO_BACK)
            {
                if (ht_p <= 0)  // back reached
                {
                    ht_phase = PH_TO_TRAGET;               // switch phase for next cycle
                    show_digit_ht(pos_to_number(9));        // show on front
                }
                else
                {
                    show_digit_ht(pos_to_number(ht_p-1));   // go back
                }
            }
            else // go to target
            {
                if (ht != ht_t) // target not reached
                {
                    show_digit_ht(pos_to_number(ht_p-1));   // go back
                }
            }
        }
        else if (ht_t < 0 && ht >= 0)  // hours ten are lit and should go off
        {
            if (ht_p <= 0)  // back reached
            {
                show_digit_ht(-1);                      // go off
            }
            else
            {
                show_digit_ht(pos_to_number(ht_p-1));   // go back
            }
        } 
        else if (ht_t >= 0 && ht < 0)  // hours ten are not lit but should go on
        {
            ht_phase = PH_TO_TRAGET;               // switch phase for next cycle
            show_digit_ht(pos_to_number(9));        // show on front
        }    
    }

#endif // END #ifdef SM_TYPE_ALL_BACK_TARGET